/**
 * Author: Dave Toth
 * Date: Fall 2005
 * 
 */

import java.net.*;      // Use sockets.
import java.io.*;       // Read and write over sockets.
import java.util.*;     // Timers, Vectors, Hashtable
import java.util.Timer; // Timers
import java.awt.*;      // Timers

////////////////////////////////////////////////////////////////////
// SCIENCE APPLICATION THREAD
////////////////////////////////////////////////////////////////////

/**
 * Functions as the main portion of the volunteer computing program.
 * Performs the computations (science application), checkpointing, 
 * and restoring from checkpoints, and creates threads to download 
 * new workunits when necessary.
 */


public class ScienceApplication extends Thread {
	
    ///////////////////////////////////////////////////////////////////////////////////
    //                                   CONSTANTS                                  //
    //////////////////////////////////////////////////////////////////////////////////


    /**
     * Constants indicating how many ffts to do to complete a workunit.
     */
    public static final Long FFTS = new Long("22053485688"); // Cel-400 24 hour task
    //public static final Long FFTS = new Long("19871285735"); // PII-350 24 hour task
    //public static final Long FFTS = new Long("18643027843"); // Cel-333 24 hour task
    public static final long FFTS_IN_WORKUNIT = FFTS.longValue(); 

    
    /**
     * Constants for fast fourier transform.
     */
    public static final float SIN_2PI_16 = (float) 0.38268343236508978;
    public static final float SIN_4PI_16 = (float) 0.707106781186547460;
    public static final float SIN_6PI_16 = (float) 0.923879532511286740;
    public static final float C_P_S_2PI_16 = (float) 1.30656296487637660;
    public static final float C_M_S_2PI_16 = (float) 0.54119610014619690;
    public static final float C_P_S_6PI_16 = (float) 1.3065629648763766;
    public static final float C_M_S_6PI_16 = (float) -0.54119610014619690;


    ///////////////////////////////////////////////////////////////////////////////////
    //                                   GLOBALS                                    //
    //////////////////////////////////////////////////////////////////////////////////

   
    /**
     * Indicates whether a thread should be paused (true) or killed (false).
     */
    public boolean threadPaused;
    
    /**
     * Indicates the task retrieval method the client is using.
     */ 
    int method;
    
    /**
     * The unique ID of the client.
     */
    String clientID;
    
    /**
     * The amount of time a task gets from download completion before it is terminated.
     */
    long delayBound;
    
    /**
     * Used to write to sockets.
     */
    private PrintWriter writer;
    
    /**
     * Used to read from sockets.
     */
    private BufferedReader reader;
    
    /**
     * The Socket to use to connect to the server.
     */
    private Socket connection;
    
    /**
     * Indicates whether it's time to create a checkpoint.
     */
    public boolean timeForCheckpoint;
    
    /**
     * The thread that keeps trace of whether it's time to create a checkpoint.
     */
    public static CheckpointThread checkpointThread;
    
    /**
     * Array that holds the numbers used for the FFTs
     */    
    float[][] input;

    /**
     * A reference to the ValidatorClient.
     */
    ValidatorClient parent;

    /**
     * Inidcates which input data set should be used for the current FFT.
     */
    int myIndex;

    /**
     * Indicates whether we were able to successfully create a checkpoint.
     */
    boolean okToProceed;

    /**
     * The number of FFTs completed.
     */
    public long fftCompleted = 0;


    /**
     * Variables for FFTs.
     */
    float[] output = new float[16];
    float temp, out0, out1, out2, out3, out4, out5, out6, out7, out8;
    float out9,out10,out11,out12,out13,out14,out15;		


    boolean brokeOut;
    
    /**
     * Constructor.
     */
    public ScienceApplication (int method, String clientID, long delayBound, ValidatorClient parent) {
	this.method = method;
	this.clientID = clientID;
	this.delayBound = delayBound;
	this.parent = parent;
	input = new float[16][9510]; 
	input[0][0] = -9999;
	brokeOut = false;
    }
    
    public void run() {

	//System.out.println("Sci App started: " + ValidatorClient.getCurrentSimulationTime());
	
	fftCompleted = 0;
	timeForCheckpoint = false;

	// This is the first time this thread has started, but it may not be the first time this task has started.
	// If there's a checkpoint file, restore from it (meaning the task was interrupted and needs to be restored).
	// Otherwise, this is the first time we've tried this task so we don't need to restore from a checkpoint.
	File f = new File("checkpoint.txt");
	if (f.exists() == true) {
	    //System.out.println("Restoring from ckpt");
	    // Restore state from checkpoint file
	    try {
		BufferedReader input = new BufferedReader(new FileReader(f));
		String line = input.readLine();
		fftCompleted = Long.parseLong(line.trim());
		input.close();
	    }
	    catch (Exception e) {
		System.out.println("Couldn't restore from checkpoint.");
	    }
	    
	    // Load the input for the FFTs from the data file.
	    loadArrays();
	}
	
	// If we fail to restore the array of FFT input, we'll need to load it again.
	// We assume success (false) to start with.
	boolean restoreFailed = false;

	//System.out.println("WHILE>>>>>>");

	// Run as long as threadPaused is true.  When threadPaused is false, the 
	// thread should be terminated.
	while (threadPaused == true) {

	    // If we've been interrupted, we should pause until we're interrupted again.
	    // At that point, we will know whether to continue executing this thread or
	    // that we should terminate this thread.
	    if ((this.isInterrupted() == true) && (threadPaused == true)) {
		pauseThread();
		if (threadPaused != false) {
		    //System.out.println("Sci App resumed: " + ValidatorClient.getCurrentSimulationTime());		    
		}

	    }
	    
	    // If this was a signal to shut down, do it.  Otherwise restart the cycle of download, compute, return result
	    if (threadPaused == false) {
		//System.out.println("Sci App killed: " + ValidatorClient.getCurrentSimulationTime());
		killCheckpointThread();
		// That's the signal that the thread should be killed.
		//System.out.println("killing and exiting");
		brokeOut = true;
		break;
	    }
	    
	    
	    // Initiate a download if necessary.  It's only necessary here if we have nothing to do now.
	    // We do the main downloads at the end of this cycle.
	    if (ValidatorClient.taskVector.size() == 0) { // There's nothing to do.
		boolean downloadSucceeded = downloadFile();
		//System.out.println("res of dl: " + downloadSucceeded + ValidatorClient.getCurrentSimulationTime());
	    }

	    // If we've been interrupted, we should pause until we're interrupted again.
	    // At that point, we will know whether to continue executing this thread or
	    // that we should terminate this thread.
	    if ((this.isInterrupted() == true) && (threadPaused == true)) {
		pauseThread();
		if (threadPaused != false) {
		    //System.out.println("Sci App resumed: " + ValidatorClient.getCurrentSimulationTime());		    
		}
	    }

	    // If this was a signal to shut down, do it.  Otherwise continue the cycle of download, compute, return result
	    if (threadPaused == false) {
		//System.out.println("Sci App killed: " + ValidatorClient.getCurrentSimulationTime());
		killCheckpointThread();
		// That's the signal that the thread should be killed.
		//System.out.println("killing and exiting");
		brokeOut = true;
		break;
	    }

	        
	    // RESTORE ARRAYS OF NUMBERS TO USE FOR FFTS
	    if (input[0][0] == -9999) { 
		//System.out.println("need to restore arrays at: " + ValidatorClient.getCurrentSimulationTime());
		restoreFailed = loadArrays();
		//System.out.println("RES of restoring arrays... " + restoreFailed + "   at: " + ValidatorClient.getCurrentSimulationTime());
	    }

	    if ((this.isInterrupted() == true) && (threadPaused == true)) {
		pauseThread();
		if (threadPaused != false) {
		    //System.out.println("Sci App resumed: " + ValidatorClient.getCurrentSimulationTime());		    
		}
	    }

	    // If this was a signal to shut down, do it.  Otherwise continue the cycle of download, compute, return result
	    if (threadPaused == false) {
		//System.out.println("Sci App killed: " + ValidatorClient.getCurrentSimulationTime());
		killCheckpointThread();
		// That's the signal that the thread should be killed.
		//System.out.println("killing and exiting");
		brokeOut = true;
		break;
	    }
	    


	    // If we were able to load the FFT input data, continue on.  Otherwise, 
	    // skip everything else so we can try to load it again as our next thing to do.
	    if (restoreFailed == false) {

		// If the current task is late, abort it.  
		long currentTime;
		currentTime = System.currentTimeMillis();
		long abortTime = ((Task) ValidatorClient.taskVector.elementAt(0)).abortTime;
		if (abortTime <= currentTime) { // Task is late...Abort
		    // Remove the checkpoint file.   
		    File f1 = new File("checkpoint.txt");
		    if (f1.exists() == true) {
			f1.delete();
		    }
		    input[0][0] = -9999; 
		    fftCompleted = 0;
		    ValidatorClient.taskVector.remove(0);
		}
		else {
		    // Do another FFT   
		    //for (int toth = 0; toth < 300000; toth ++) {
			doFFT();
		    //}
		    //System.out.println("FFT COMP: " + fftCompleted);
		    
		    // Do a checkpoint if necessary.   
		    if (timeForCheckpoint == true) {
			
			// We assume it takes 10 sec to create a checkpoint
			okToProceed = true;
			try {
			    Thread.sleep(10000);
			}
			catch (InterruptedException exc) {
			    // We were unable to take the time required to create a checkpoint so we don't do it.
			    checkpointThread = null;
			    okToProceed = false;
			    
			    // If this was a signal to shut down, do it.  Otherwise continue the cycle of download, compute, return result
			    if (threadPaused == false) {
				killCheckpointThread();
				// That's the signal that the thread should be killed.
				//System.out.println("killing and exiting");
				brokeOut = true;
				break;
			    }
			    else {
				// That's the signal to pause the thread.
				pauseThread();
				if (threadPaused != false) {
				    //System.out.println("Sci App resumed: " + ValidatorClient.getCurrentSimulationTime());		    
				}
			    }

			    // If this was a signal to shut down, do it.  Otherwise continue the cycle of download, compute, return result
			    if (threadPaused == false) {
				//System.out.println("Sci App killed: " + ValidatorClient.getCurrentSimulationTime());
				killCheckpointThread();
				// That's the signal that the thread should be killed.
				//System.out.println("killing and exiting");
				brokeOut = true;
				break;
			    }	
			}
			
			if (okToProceed == true) {
			    // We were able to use the 10 sec required to create a checkpoint so we do it.
			    
			    // Remove the old checkpoint file.   
			    File f1 = new File("checkpoint.txt");
			    if (f1.exists() == true) {
				f1.delete();
			    }
			    
			    // Create the new checkpoint file.
			    File f2 = new File("checkpoint.txt");
			    try {
				FileWriter writer = new FileWriter(f2);
				writer.write(fftCompleted + "\n");
				// Close the checkpoint file.
				writer.close();
				//System.out.print("#");
			    }
			    catch (IOException e) {
				System.out.println("Failed to create checkpoint");
			    }
			    
			    if (checkpointThread != null) {
				try {
				    checkpointThread.setRunning(false);
				    checkpointThread.interrupt();
				}
				catch (NullPointerException npee) {
				// Do nothing - everything is already taken care of.
				}
			    }
			    checkpointThread = null;
			    checkpointThread = new CheckpointThread(this);
			    checkpointThread.start();
			    timeForCheckpoint = false;
			}
		    }
		    
		    
		    // Initiate a download if necessary for download early   
		    if ((fftCompleted >= (.95 * FFTS_IN_WORKUNIT)) && (ValidatorClient.threadVector.size() <= 1) && (ValidatorClient.taskVector.size() <= 1 ) && (method == ValidatorClient.DOWNLOAD_EARLY)) {
			DownloadThread dlthread = new DownloadThread(this, clientID, delayBound);
			dlthread.threadPaused = true;
			ValidatorClient.threadVector.add(dlthread);
			dlthread.start();
		    }
		    
		    
		    
		    // If done, return the result and remove checkpoint file.   
		    if (fftCompleted >= FFTS_IN_WORKUNIT) {
			
			// Finalize this task by returning the result to the server, ensuring 
			// no more cheeckpoints get created for this task, and removing all
			// the old checkpoints.
			finalizeTask();
		    }
		    
		    
		    
		    // Initiate download if finished workunit and need to download.   
		    int numToDownload = needToDownload(method);
		    for (int x = 0; x < numToDownload; x ++) {
			DownloadThread dlthread = new DownloadThread(this, clientID, delayBound);
			dlthread.threadPaused = true;
			ValidatorClient.threadVector.add(dlthread);
			dlthread.start();
		    }
		}
	    }
	} // END WHILE
	if (brokeOut == false) {
	    //System.out.println("&&&& Sci App killed" + ValidatorClient.getCurrentSimulationTime());
	    killCheckpointThread();
	}
	ValidatorClient.threadVector.remove(this);
	//System.out.println("Thread dead - exiting program");
    }
    
    /**
     * Returns the result of the task to the server, ensuring 
     * no more cheeckpoints get created for this task, and 
     * removing all the old checkpoints.
     */
    public void finalizeTask() {
	//Return the result.
	Task task = (Task) ValidatorClient.taskVector.remove/*elementAt*/(0);
	returnResult(task.taskID);
	parent.incrementTasksCompleted();
	System.out.println("Task Completed: " + parent.getTasksCompleted());
	
	// Make sure that no more checkpoint files are created for this task.
	if (checkpointThread != null) {
	    try {
		checkpointThread.setRunning(false);
		checkpointThread.interrupt();
		checkpointThread = null;
	    }
	    catch (NullPointerException npe) {
		// Do Nothing.  The thread is already dead.
	    }
	}
	
	// Remove the checkpoint file.   
	File f1 = new File("checkpoint.txt");
	if (f1.exists() == true) {
	    f1.delete();
	}
	fftCompleted = 0;
	input[0][0] = -9999; 
    }

    /**
     * Loads the inputs for the FFTs into the array.
     * @return Returns whether this task was successfully completed
     */
    public boolean loadArrays() {
	//System.out.println("@@ loading arrays at: " + ValidatorClient.getCurrentSimulationTime());
	try {
	    // Open the file
	    File dataFile = new File(((Task) ValidatorClient.taskVector.elementAt(0)).fileName);
	    BufferedReader inputReader = new BufferedReader(new FileReader(dataFile));
	    String line;
	    
	    // Read in all 9510 lines, storing the values in the array.
	    for (int s = 0; s < 9510; s ++) {
		line = inputReader.readLine();
		String data[] = line.split(",");
		if (data.length < 16) {
		    System.out.println("data length error in load arrays() - line: " + s + "only " + data.length + " blocks");
		}
		for (int inner = 0; inner < 16; inner ++) {
		    input[inner][s] = Float.parseFloat(data[inner]);
		}
	    }
	    
	    // Close the file.
	    inputReader.close();
	    
	}
	catch (Exception e) {
	    //interrupted = true;
	    System.out.println("Interruption while restoring data file " + ValidatorClient.getCurrentSimulationTime() + " - " + e.toString());
	    input[0][0] = -9999; 
	    // If this task was interrupted or failed for any reason, we 
	    // need to try again, so return true for failure.

	    // If this was an interrupted exception, then we're supposed to pause and we'll return the failure when we resume
	    if (e instanceof InterruptedException) {
		
		pauseThread();
	    }
	    
	    return true;
	}
	// We loaded the arrays successfully. 
	// Start the checkpoint thread if necessary
	if (checkpointThread == null) {
	    checkpointThread = new CheckpointThread(this);
	    checkpointThread.start();
	} 
	else {
	    killCheckpointThread();
	    checkpointThread = null;
	    checkpointThread = new CheckpointThread(this);
	    checkpointThread.start();
	}
	
	return false;
    }

    /**
     * Does an FFT
     */
    public void doFFT() {
	myIndex = (int) (fftCompleted % 9510);
	
	out0=input[0][myIndex]+input[8][myIndex]; // output[0 through 7] is the data
	out1=input[1][myIndex]+input[9][myIndex]; // that we take the 8 point real 
	out2=input[2][myIndex]+input[10][myIndex]; //FFT of. 
	out3=input[3][myIndex]+input[11][myIndex];
	out4=input[4][myIndex]+input[12][myIndex];
	out5=input[5][myIndex]+input[13][myIndex];
	out6=input[6][myIndex]+input[14][myIndex];
	out7=input[7][myIndex]+input[15][myIndex];
	
	out8=input[0][myIndex]-input[8][myIndex];   /* inputs 8,9,10,11 are */
	out9=input[1][myIndex]-input[9][myIndex];   /* the Real part of the */
	out10=input[2][myIndex]-input[10][myIndex]; /* 4 point Complex FFT inputs.*/
	out11=input[3][myIndex]-input[11][myIndex]; 
	out12=input[12][myIndex]-input[4][myIndex]; /* outputs 12,13,14,15 are */
	out13=input[13][myIndex]-input[5][myIndex]; /* the Imaginary pars of  */
	out14=input[14][myIndex]-input[6][myIndex]; // the 4 point Complex FFT inputs
	out15=input[15][myIndex]-input[7][myIndex];
	
	/*First we do the "twiddle factor" multiplies for the 4 
	  point CFFT 
	  Note that we use the following handy trick for doing a 
	  complex multiply:  (e+jf)=(a+jb)*(c+jd) */
	/*   e=(a-b)*d + a*(c-d)   and    f=(a-b)*d + b*(c+d)  */
	
	/* C_M_S_2PI/16=cos(2pi/16)-sin(2pi/16) when replaced by 
	   macroexpansion */
	/* C_P_S_2PI/16=cos(2pi/16)+sin(2pi/16) when replaced by 
	   macroexpansion */
	// (SIN_2PI_16)=sin(2pi/16) when replaced by macroexpansion
	temp=(out13-out9)*(SIN_2PI_16); 
	out9=out9*(C_P_S_2PI_16)+temp; 
	out13=out13*(C_M_S_2PI_16)+temp;
	
	out14*=(SIN_4PI_16);
	out10*=(SIN_4PI_16);
	out14=out14-out10;
	out10=out14+out10+out10;
	
	temp=(out15-out11)*(SIN_6PI_16);
	out11=out11*(C_P_S_6PI_16)+temp;
	out15=out15*(C_M_S_6PI_16)+temp;
	
	// The following are the first set of two point butterfiles
	// for the 4 point CFFT 
	
	out8+=out10;
	out10=out8-out10-out10;
	
	out12+=out14;
	out14=out12-out14-out14;
	
	out9+=out11;
	out11=out9-out11-out11;
	
	out13+=out15;
	out15=out13-out15-out15;
	
	/*The followin are the final set of two point butterflies */
	output[1]=out8+out9;
	output[7]=out8-out9;
	
	output[9]=out12+out13;
	output[15]=out13-out12;
	
	output[5]=out10+out15;    /* implicit multiplies by */
	output[13]=out14-out11;   /* a twiddle factor of -j */     
	output[3]=out10-out15;    /* implicit multiplies by */
	output[11]=-out14-out11;  /* a twiddle factor of -j */
	
	
	/* What follows is the 8-point FFT of points output[0-7] */
	/* This 8-point FFT is basically a Decimation in Frequency
	   FFT where we take advantage of the fact that the 
	   initial data is real*/
	
	/* First set of 2-point butterflies */
	
	out0=out0+out4;
	out4=out0-out4-out4;
	out1=out1+out5;
	out5=out1-out5-out5;
	out2+=out6;
	out6=out2-out6-out6;
	out3+=out7;
	out7=out3-out7-out7;
	
	/* Computations to find X[0], X[4], X[6] */
	
	output[0]=out0+out2;
	output[4]=out0-out2;
	out1+=out3;
	output[12]=out3+out3-out1;
	
	output[0]+=out1;  /* Real Part of X[0] */
	output[8]=output[0]-out1-out1; /*Real Part of X[4] */
	/* out2 = Real Part of X[6] */
	/* out3 = Imag Part of X[6] */
		
	/* Computations to find X[5], X[7] */
	
	out5*=SIN_4PI_16;
	out7*=SIN_4PI_16;
	out5=out5-out7;
	out7=out5+out7+out7;
	
	output[14]=out6-out7; /* Imag Part of X[5] */
	output[2]=out5+out4; /* Real Part of X[7] */
	output[6]=out4-out5; /*Real Part of X[5] */
	output[10]=-out7-out6; /* Imag Part of X[7] */
	
	fftCompleted ++;
    }

    /**
     * Downloads a workunit.
     * @return Returns false if the download failed or was interrupted.
     */
    public boolean downloadFile() {
	if (ValidatorClient.threadVector.size() <= 1) { // We're not downloading already
	    // Initiate Download.  
	    DownloadThread dlthread = new DownloadThread(this, clientID, delayBound);
	    dlthread.threadPaused = true;
	    ValidatorClient.threadVector.add(dlthread);
	    dlthread.start();
	}
	
	// Wait for the download to complete.  We can't do anything useful until it does.  
	boolean downloadComplete = false;
	if (checkpointThread != null) {
	    try {
		checkpointThread.setRunning(false);
		checkpointThread.interrupt();
		checkpointThread = null;
	    }
	    catch (NullPointerException npe) {
		// Do Nothing.  The thread is already dead.
	    }
	}
	while (downloadComplete == false) {
	    try {
		Thread.sleep(250);
	    }
	    catch (InterruptedException inte) {
		//System.out.println("DL thread in DLFILE() interrupted at: " + ValidatorClient.getCurrentSimulationTime());
		if (threadPaused == false) {
		    killCheckpointThread();
		    // That's the signal that the thread should be killed.
		    return false;
		}
		else {
		    pauseThread();
		    if (threadPaused == false) {
			return false;
		    }
		}
	    }	

	    if (this.isInterrupted() == true) {
		//System.out.println("DL thread in DLFILE() interrupted at: " + ValidatorClient.getCurrentSimulationTime());
		if (threadPaused == false) {
		    killCheckpointThread();
		    // That's the signal that the thread should be killed.
		    return false;
		}
		else {
		    pauseThread();
		    if (threadPaused == false) {
			return false;
		    }
		}
	    }
	    if (ValidatorClient.taskVector.size() > 0) {
		downloadComplete = true;
	    }
	}
	 	
	// We made it through this method without being interrupted or failing to download.
	//System.out.println("DL thread in DLFILE() completed at: " + ValidatorClient.getCurrentSimulationTime());
	return true;
    }

    /**
     * Instructs the checkpoint Thread to terminate.
     */
    public void killCheckpointThread() {
	//System.out.print("KILLING CKPT TH FROM SCI APP");
	if (checkpointThread != null) {
	    try {
		checkpointThread.setRunning(false);
		checkpointThread.interrupt();
		checkpointThread = null;
	    }
	    catch (NullPointerException npe) {
		// Do Nothing.  The thread is already dead.
		checkpointThread = null;
	    }
	}
    }


    /**
     * Pause the execution of the science application until
     *  another interrupt is received.
     */
    public void pauseThread() {
	//System.out.print("pausing: " + ValidatorClient.getCurrentSimulationTime() + "    signaling ckpt th...");
	// That's the signal to pause the thread.
	if (checkpointThread != null) {
	    try {
		//System.out.println("sending sig to ckpt");
		checkpointThread.interrupt();
	    }
	    catch (NullPointerException npe) {
		//System.out.println("can't send sig to ckpt - null");
		// Do Nothing.  The thread is already dead.
	    }
	}

	if (Thread.interrupted()) {/*System.out.println("---------------------- squelching");*/}

	boolean ready = false;
	while (!ready) {
	    try {
		// Sleep for 14 days.  We'll be woken up before then.
		Thread.sleep(1209600000); 
	    }
	    catch (InterruptedException ie) {
		ready = true;
		//System.out.print("DEAL WITH CKPT TH... ");
		if (checkpointThread == null) {
		    //System.out.println("^^^^^^^^^^^^^ The ckpt thread was dead - should not be");
		    checkpointThread = new CheckpointThread(this);
		    checkpointThread.start();
		}
		else {
		    //System.out.println("waking ckpt thread");
		    checkpointThread.interrupt();
		}
	    }
	}
    }

    /**
     * Determines whether the client needs to download another task based on the task retrieval method used.
     * @param method - The task retrieval method.
     * @return Returns how many tasks the client needs to download.
     */
    public int needToDownload(int method) {
	
	float cpuPercentAvailable;
	float timeAvailableToProcess;
	int maxTasks;
	int timeAvailable;
	
	switch (method) {
	case ValidatorClient.BUFFER_NONE:
	    if ((ValidatorClient.taskVector.size() == 0) && (ValidatorClient.threadVector.size() < 2)) {
		return 1;
	    }
	    else {
		return 0;
	    }
	    
	case ValidatorClient.DOWNLOAD_EARLY:
	    // All of the cases for this method are handled without calling this function. 
	    return 0;
	    
	case ValidatorClient.BUFFER_1_TASK:
	    if (ValidatorClient.taskVector.size() >= 2) {
		return 0;
	    }
	    else {
		if (ValidatorClient.taskVector.size() == 1) {
		    if (ValidatorClient.threadVector.size() >= 2) {
			return 0;
		    }
		    else {
			return 1;
		    }
		}
		else {
		    if (ValidatorClient.threadVector.size() == 1) {
			return 2;
		    }
		    else if (ValidatorClient.threadVector.size() == 2) {
			return 1;
		    }
		    else {
			return 0;
		    }
		}
	    }
	    
	    // We assume that a task is 24 hours for these next cases.
	    // We also assume if there's at least one task buffered, there's no need to refill the buffer.
	    // Additionally, if there's at least 1 task being worked on and at least one task being downloaded now,
	    // the buffer is being refilled already so we don't need to do anything else.
	case ValidatorClient.BUFFER_1_DAY:
	    if (ValidatorClient.taskVector.size() >= 2) {
		return 0;
	    }
	    if ((ValidatorClient.taskVector.size() == 1) && (ValidatorClient.threadVector.size() >= 2)) {
		return 0;
	    }
	    
	    cpuPercentAvailable = getPercentCpuAvailable(ValidatorClient.traceFile);
	    timeAvailableToProcess = (cpuPercentAvailable/100) * ((float) delayBound);
	    timeAvailable = (int) timeAvailableToProcess;
	    timeAvailable = timeAvailable - (86400 * (Math.max(0, (ValidatorClient.taskVector.size() - 1)) + (ValidatorClient.threadVector.size() - 1)));
	    
	    maxTasks = (int) (timeAvailable/86400);
	    return Math.min(1,maxTasks);
	    
	case ValidatorClient.BUFFER_3_DAYS:
	    if (ValidatorClient.taskVector.size() >= 2) {
		return 0;
	    }
	    if ((ValidatorClient.taskVector.size() == 1) && (ValidatorClient.threadVector.size() >= 2)) {
		return 0;
	    }
	    
	    cpuPercentAvailable = getPercentCpuAvailable(ValidatorClient.traceFile);
	    timeAvailableToProcess = (cpuPercentAvailable/100) * ((float) delayBound);
	    timeAvailable = (int) timeAvailableToProcess;
	    timeAvailable = timeAvailable - (86400 * (Math.max(0, (ValidatorClient.taskVector.size() - 1)) + (ValidatorClient.threadVector.size() - 1)));
	    
	    maxTasks = (int) (timeAvailable/86400);
	    return Math.min(3,maxTasks);
	    
	case ValidatorClient.BUFFER_7_DAYS:
	    if (ValidatorClient.taskVector.size() >= 2) {
		return 0;
	    }
	    if ((ValidatorClient.taskVector.size() == 1) && (ValidatorClient.threadVector.size() >= 2)) {
		return 0;
	    }
	    
	    cpuPercentAvailable = getPercentCpuAvailable(ValidatorClient.traceFile);
	    timeAvailableToProcess = (cpuPercentAvailable/100) * ((float) delayBound);
	    timeAvailable = (int) timeAvailableToProcess;
	    timeAvailable = timeAvailable - (86400 * (Math.max(0, (ValidatorClient.taskVector.size() - 1)) + (ValidatorClient.threadVector.size() - 1)));
	    
	    maxTasks = (int) (timeAvailable/86400);
	    return Math.min(7,maxTasks);
	    
	case ValidatorClient.BUFFER_14_DAYS:
	    if (ValidatorClient.taskVector.size() >= 2) {
		return 0;
	    }
	    if ((ValidatorClient.taskVector.size() == 1) && (ValidatorClient.threadVector.size() >= 2)) {
		return 0;
	    }
	    
	    cpuPercentAvailable = getPercentCpuAvailable(ValidatorClient.traceFile);
	    timeAvailableToProcess = (cpuPercentAvailable/100) * ((float) delayBound);
	    timeAvailable = (int) timeAvailableToProcess;
	    timeAvailable = timeAvailable - (86400 * (Math.max(0, (ValidatorClient.taskVector.size() - 1)) + (ValidatorClient.threadVector.size() - 1)));
	    
	    maxTasks = (int) (timeAvailable/86400);
	    return Math.min(14,maxTasks);
	} 
	return 0;
    }
    
        
    public void bufferWorkunit(Task task) {
	ValidatorClient.taskVector.add(task);
    }
    
    
    /**
     * Returns the percent of cpu time (as a float) that a computer is avaiable
     * for volunteer computing.
     * @return - Returns the percent of cpu time (as a float) that a computer
     *           is avaiable for volunteer computing.
     */
    float getPercentCpuAvailable(String trace) {
	Float val = (Float) ValidatorClient.cpuAvailabilityTable.get(trace);
	float f = val.floatValue();
	return f;
    }
    
    /**
     * Returns the result of the work unit to the server.  In a real system
     * we might have a whole set of results, so we'd have to modify this a bit.
     * @param - The task ID of the work unit.
     * @return - Returns the status of sending the computation result to the
     *           server.
     */
    public boolean returnResult(String workUnitID) {
	if (connectToServer()) {
	    writer.write(ValidatorClient.RETURN_RESULT + "-" + clientID + "-" + workUnitID + "=" + 0 + "=" + 0 + "\n");
	    disconnectFromServer();
	    return true;
	}
	return false;
    }
    
    
    /** 
     * Establish a connection to the server and return the result.
     * @return Returns OK if successful, FAILED otherwise.
     */
    public boolean connectToServer() {
	try {
	    connection = new Socket("192.168.2.21", 19021);
	    writer = new PrintWriter(connection.getOutputStream());
	    reader = new BufferedReader(new InputStreamReader(connection.getInputStream()));
	    return true;
	}
	catch (Exception e) {
	    return false;
	}
    }
    
    
    /**
     * Disconnect from the server.
     */
    public void disconnectFromServer() {
	try {
	    writer.close();
	    reader.close();
	    connection.close();
	}
	catch (Exception e) {
	    // Do nothing.  We're probably already disconnected.
	}
    } 
}

