/**
 * Author: Dave Toth
 * Date: Fall 2005
 * 
 * This class acts as a stripped-down public-resource computing server.  It
 * runs on a server along with a web server.  The program receives connections
 * requesting work and responds with a file name based on a parameter passed
 * to it by the client program.  The file name indicates which data file the 
 * client should download from the webserver to use for its computation.  
 * This program also ensures that the client responds with a correct answer 
 * to the computation.
 */

import java.io.*;  // Read and write over sockets
import java.net.*; // Use sockets.
import java.sql.*; // Use Databases

public class Server {

    //////////////////////////////////////////////////////////////////////
    //                       Global Variables                           //
    //////////////////////////////////////////////////////////////////////

    /**
     * Used to receive incoming connections.
     */
    ServerSocket incomingSocket;

    /**
     * Used to communicate with the clients.
     */
    Socket connection;

    /**
     * Indicates that the program should continue running, accepting and 
     * servicing connects.  When we get an exit signal from a client, 
     * we'll clean up everything and shut down.
     */
    boolean keepRunning;

    /**
     * The ID number of the next workunit to assign.
     */
    int nextWorkunitID;


    //////////////////////////////////////////////////////////////////////
    //                          Methods                                 //
    //////////////////////////////////////////////////////////////////////


    /**
     * The main method.  Create a new instance of the Server class and 
     * hand control over to it.
     * @param args - The array of command line arguments passed to the JVM.
     */
    public static void main(String[] args) {
	Server server = new Server();
    }


    /**
     * Constructor.  Open the ServerSocket to listen for incoming connections
     * and spawn a new Thread to handle each new connection.  If there's an 
     * exception setting up the ServerSocket, something is grossly wrong and 
     * we can't continue, so exit.  If there's an exception establishing an 
     * individual connection, we'll let that connection fail and keep 
     * servicing any others that come in.
     */
    public Server() {
	keepRunning = true;
	nextWorkunitID = 2;//1;

	System.out.println("Server Started - Connecting to Database...\n");

	/*
	try {
	    Class.forName("sun.jdbc.odbc.JdbcOdbcDriver");
	    Connection connection = DriverManager.getConnection("jdbc:odbc:daves_data_source");
	    System.out.println("Connection Established");
	    connection.close();
	    System.out.println("Connection Closed");
	}
	catch (ClassNotFoundException cnfe) {
	    System.out.println("ClassNotFoundException");
	}
	catch (SQLException sqle) {
	    System.out.println("SQLException");
	}

	System.out.println("Established Connection to Database.  Ready to accept connections from volunteer computing clients\n");
	*/

	try {
	    incomingSocket = new ServerSocket(19021);

	    // Just run forever accepting connections until we get an exit message.
	    while (keepRunning == true) {
		connection = new Socket();
		try {
		    connection = incomingSocket.accept();
		    ConnectionManager manager = new ConnectionManager(connection, this);
		    manager.start();
		}
		catch (Exception e) {
		    // Print the error so we see it, but continue the program.
		    System.out.println("Error with sockets");
		}
	    }
	}
	catch (Exception e) {
	    // Setting up the ServerSocket failed.  We can't recover from this
	    // so just exit.
	    System.out.println("hey ey bye");
	    System.exit(0);
	}

	// We received the shutdown message, so let's wait a minute to let everything
	// finish and then shutdown all the connections and exit.
	System.out.println("Received shutdown message... shutting down in 1 minute.");
	try {
	    Thread.sleep(10000); // This is 10 seconds, not 60!!!
	}
	catch (InterruptedException ie) {
	    // Do nothing.
	}
	System.exit(0);
    }


    synchronized int getNextWorkunitID() {
	nextWorkunitID ++;
	return nextWorkunitID - 1;
    }



    //////////////////////////////////////////////////////////////////////
    //                            Inner Classes                         //
    //////////////////////////////////////////////////////////////////////

    /**
     * An instance of this class is created to handle the work for each new 
     * connection.  It listens to the socket and responds to the client, 
     * telling the client what file to download for the client's computation
     * or recording the client's answer to a computation and validating it.
     * Each instance of this class is its own Thread.
     */
    class ConnectionManager extends Thread {
	BufferedReader reader;
	PrintWriter writer;
	Socket socket;
	Server parent;

	/**
	 * Constructor.  Create the readers and writers to communicate with a
	 * client over the socket.
	 * @param socket - The socket used to communicate with the client.
	 * @param parent - The server that spawned this thread.
	 */
	public ConnectionManager(Socket socket, Server parent) {
	    this.socket = socket;
	    this.parent = parent;
	    try {
		reader = new BufferedReader(new 
		    InputStreamReader(socket.getInputStream()));
		writer = new PrintWriter(socket.getOutputStream());
	    }
	    catch (Exception e) {
		System.out.println("HELP");
		// TODO - DO SOMETHING HERE
	    }
	}


	/**
	 * The code to run once an instance of the class has been created.
	 */
	public void run() {
	    String msg;
	    boolean active = true;
	    while (active == true) {
		// Read any messages from the client.
		try {
		    msg = reader.readLine();

		    // If the client is requesting work...
		    if (msg.indexOf(ValidatorClient.REQUEST_WORK) >= 0) {

			System.out.println(msg);

			// Send the appropriate file name based on the two 
			// numbers in the request.
			String command = (msg.substring(msg.indexOf(" "))).trim();
			int id = getNextWorkunitID();
			System.out.println("Sending: " + id + "_" + "work" + id + ".txt");
			writer.println(id + "_" + "work" + id + ".txt");

			// Send a compuation for the client to do.
			writer.flush();

			// Disconnect.  
			try {
			    reader.close();
			    writer.close();
			    socket.close();
			}
			catch (Exception e) {
			    // There's nothing we can do to recover from this
			    // but there's probably not a need to do anything
			    // so just report the error and keep going.
			    System.out.println("Error closing socket.");
			}
			active = false;
		    }

		    // The client is returning an answer.
		    else if (msg.indexOf(ValidatorClient.RETURN_RESULT) >= 0) {

			long returnedTime = System.currentTimeMillis();

			System.out.println(msg);

			// Disconnect.
			try {
			    reader.close();
			    writer.close();
			    socket.close();			    
			}
			catch (Exception e) {
			    // There's nothing we can do to recover from this
			    // but there's probably not a need to do anything
			    // so just report the error and keep going.
			    System.out.println("Error closing socket.");
			}
			// Now process the message.
			// Find the workunitID of the workunit that has been completed.

			String clientID = msg.substring(1 + msg.indexOf("-"), msg.lastIndexOf("-"));
			String workUnitID = msg.substring(1 + msg.lastIndexOf("-"), msg.indexOf("="));
			String answer = msg.substring(1 + msg.lastIndexOf("="));
			System.out.println(clientID + "X" + workUnitID + "X" + answer + "X");

			boolean done = false;
			while (!done) {
			    // Make a connection to the database and update the record.
			    try {
				///REMClass.forName("sun.jdbc.odbc.JdbcOdbcDriver");
				///REMConnection connection = DriverManager.getConnection("jdbc:odbc:daves_data_source");

				// Update the Answer
				///REMStatement statement = connection.createStatement();
				String sql = "UPDATE validation_results set Answer = " + answer + " where TaskID = " + workUnitID;
				System.out.println("SQL: " + sql);
				///REMint result = statement.executeUpdate(sql);
				///REMSystem.out.println("RES OF SQL UPDATE: " + result);

				// Find out what the TimeDue value was for this workunit and determine whether it was completed on time.
				///REMstatement = connection.createStatement();
				sql = "select TimeDue from validation_results where TaskID = " + workUnitID;
				System.out.println("SQL: " + sql);
				///REMResultSet resSet = statement.executeQuery(sql);
				long dueDate = 0;

				///REMwhile (resSet.next()) {
				///REM    dueDate = resSet.getLong("TimeDue");
				///REM    System.out.println("\n\n DUE AT... " + dueDate);
				///REM}

				
				// Update the CompletedOnTime
				///REMstatement = connection.createStatement();
				if (dueDate >= returnedTime) {
				    sql = "UPDATE validation_results set CompletedOnTime = 1 where TaskID = " + workUnitID;
				}
				else {
				    sql = "UPDATE validation_results set CompletedOnTime = 0 where TaskID = " + workUnitID;
				}
				System.out.println("SQL: " + sql);
				///REMresult = statement.executeUpdate(sql);
				///REMSystem.out.println("RES OF SQL UPDATE: " + result);

				// Update the TimeReturned
				///REMstatement = connection.createStatement();
				sql = "UPDATE validation_results set TimeReturned = " + Long.toString(returnedTime) + " where TaskID = " + workUnitID;
				System.out.println("SQL: " + sql);
				///REMresult = statement.executeUpdate(sql);
				///REMSystem.out.println("RES OF SQL UPDATE: " + result);

				///REMconnection.close();
				done = true;	   
				Class.forName("sun.jdbc.odbc.JdbcOdbcDriver"); // TODO REMOVE THIS LINE  
			    }
			    
			    catch (ClassNotFoundException cnfe) {
				System.out.println("ClassNotFoundException");
			    }
			    /*
			    catch (SQLException sqle) {
				System.out.println("SQLException");
			    }
			    */
			}

			active = false;
		    }

		    // The client is done downloading a file.
		    else if (msg.indexOf(ValidatorClient.ACKNOWLEDGE_DOWNLOAD) >= 0) {

			System.out.println("receiving ACK!!!!!!!!!!!!!!!!!!!!!");
			long ackTime = System.currentTimeMillis();

			System.out.println(msg);

			// Disconnect.
			try {
			    reader.close();
			    writer.close();
			    socket.close();			    
			}
			catch (Exception e) {
			    // There's nothing we can do to recover from this
			    // but there's probably not a need to do anything
			    // so just report the error and keep going.
			    System.out.println("Error closing socket.");
			}
			// Now process the message.

			String clientID = msg.substring(1 + msg.indexOf("-"), msg.lastIndexOf("-"));
			String workUnitID = msg.substring(1 + msg.lastIndexOf("-"), msg.indexOf("="));
			long delayBound = Long.parseLong(msg.substring(1 + msg.lastIndexOf("=")));
			
			System.out.println(clientID + "X" + workUnitID + "X" + delayBound + "X");

			// Find the workunitID of the workunit that has been downloaded.

			boolean done = false;
			while (!done) {
			    // Make a connection to the database and update the record.
			    try {	
				///REMClass.forName("sun.jdbc.odbc.JdbcOdbcDriver");
				///REMConnection connection = DriverManager.getConnection("jdbc:odbc:daves_data_source");
				///REMStatement statement = connection.createStatement();
				String sql = "INSERT INTO validation_results" + "(ClientID, TaskID, CompletedOnTime, TimeAssigned, TimeDue, TimeReturned, Answer, Aux1, Aux2)" + "VALUES (" + clientID + "," + workUnitID + "," + "null," + Long.toString(ackTime) +"," + Long.toString(ackTime + delayBound) + ",null,null, null,null)";
				System.out.println("SQL: " + sql);
				///REMint result = statement.executeUpdate(sql);
				///REMSystem.out.println("RES OF SQL INSERT: " + result);
				///REMconnection.close();
				done = true;	 
				Class.forName("sun.jdbc.odbc.JdbcOdbcDriver"); // TODO REMOVE THIS LINE  
			    }
			    catch (ClassNotFoundException cnfe) {
				System.out.println("ClassNotFoundException" + cnfe);
			    }
			    /*
			    catch (SQLException sqle) {
				System.out.println("SQLException" + sqle);
			    }
			    */
			    //done = true; // REMOVE THIS (TODO)
			}
			active = false;
		    }

		    // We were told by a client to shut down.
		    else if (msg.indexOf(ValidatorClient.SHUTDOWN_SERVER) >= 0) {
			System.out.println("Received shutdown cmd");
			// Disconnect.
			try {
			    reader.close();
			    writer.close();
			    socket.close();			    
			}
			catch (Exception e) {
			    // There's nothing we can do to recover from this
			    // but there's probably not a need to do anything
			    // so just report the error and keep going.
			    System.out.println("Error closing socket.");
			}
			// Now process the message.
			parent.keepRunning = false;
			active = false;
			System.out.println("finished receiving shutdown cmd");
		    }

		    // There was a message that neither asked for work or 
		    // returned a result or acknowledged a download or 
		    // said to exit.  Just print the message so we can
		    // examine it later. 
		    else { 
			System.out.println("Unexpected Message: " + msg);
			// Disconnect.
			try {
			    reader.close();
			    writer.close();
			    socket.close();			    
			}
			catch (Exception e) {
			    // There's nothing we can do to recover from this
			    // but there's probably not a need to do anything
			    // so just report the error and keep going.
			    System.out.println("Error closing socket.");
			}
			active = false;
		    }
		}
		// There was an exception communicating with the client which
		// means we've lost the connection to it.  Set the variable
		// active to false so we break out of the loop.
		catch (Exception e) {
		    System.out.println("Lost connection to client");
		    active = false;
		}
	    }
	}
    }
}  
